import torch
import torch.nn as nn
from timm.layers import DropPath

from torch.autograd import Function
from torch.cuda.amp import custom_fwd, custom_bwd

try:
    from torch.utils.cpp_extension import load

    print("[Start compiling NAT]")
    nattenav_cuda = load(
        "nattenav_cuda",
        ["core/cuda/nattenav_cuda.cpp", "core/cuda/nattenav_cuda_kernel.cu"],
        verbose=True,
    )
    print("[Finished 1/2]")
    nattenqkrpb_cuda = load(
        "nattenqkrpb_cuda",
        ["core/cuda/nattenqkrpb_cuda.cpp", "core/cuda/nattenqkrpb_cuda_kernel.cu"],
        verbose=False,
    )
    print("[Finished 2/2]")
except:
    print("Failed to load nat cuda")
    exit()


class NATTENAVFunction(Function):
    @staticmethod
    @custom_fwd(cast_inputs=torch.float16)
    def forward(ctx, attn, value):
        attn = attn.contiguous()
        value = value.contiguous()
        out = nattenav_cuda.forward(attn, value)[0]
        ctx.save_for_backward(attn, value)
        return out

    @staticmethod
    @custom_bwd
    def backward(ctx, grad_out):
        outputs = nattenav_cuda.backward(grad_out.contiguous(), *ctx.saved_variables)
        d_attn, d_value = outputs
        return d_attn, d_value


class NATTENQKRPBFunction(Function):
    @staticmethod
    @custom_fwd(cast_inputs=torch.float16)
    def forward(ctx, query, key, rpb):
        query = query.contiguous()
        key = key.contiguous()
        attn = nattenqkrpb_cuda.forward(query, key, rpb.contiguous())[0]
        ctx.save_for_backward(query, key)
        return attn

    @staticmethod
    @custom_bwd
    def backward(ctx, grad_out):
        outputs = nattenqkrpb_cuda.backward(grad_out.contiguous(), *ctx.saved_variables)
        d_query, d_key, d_rpb = outputs
        return d_query, d_key, d_rpb


class selfattentionlayer_nat(nn.Module):
    def __init__(
        self,
        dropout,
        droppath,
        attn_dim,
        expand_factor,
        encoder_latent_dim,
        vert_c_dim,
        cost_latent_dim,
    ):
        super(selfattentionlayer_nat, self).__init__()
        kernel_size = 11
        qk_dim = attn_dim
        self.num_heads = qk_dim // 16
        self.expand_factor = expand_factor

        self.context_proj = nn.Linear(encoder_latent_dim, vert_c_dim, bias=True)
        self.context_norm = nn.LayerNorm(encoder_latent_dim)
        self.norm1 = nn.LayerNorm(cost_latent_dim + vert_c_dim)
        self.norm1_v = nn.LayerNorm(cost_latent_dim)
        self.norm2 = nn.LayerNorm(cost_latent_dim)

        self.q, self.k, self.v = (
            nn.Linear(cost_latent_dim + vert_c_dim, qk_dim, bias=True),
            nn.Linear(cost_latent_dim + vert_c_dim, qk_dim, bias=True),
            nn.Linear(cost_latent_dim, qk_dim, bias=True),
        )

        self.proj = nn.Linear(attn_dim + cost_latent_dim, cost_latent_dim, bias=True)

        self.drop_path = DropPath(droppath) if droppath > 0.0 else nn.Identity()

        self.ffn = nn.Sequential(
            nn.Linear(cost_latent_dim, cost_latent_dim * expand_factor),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(cost_latent_dim * expand_factor, cost_latent_dim),
            nn.Dropout(dropout),
        )

        self.dim = qk_dim
        self.rpb = nn.Parameter(
            torch.zeros(self.num_heads, (2 * kernel_size - 1), (2 * kernel_size - 1))
        )
        self.scale = 16**-0.5

    def forward(self, x, size, context):
        B, HW, C = x.shape
        B_context, C_context, H, W = context.shape

        short_cut = x
        context = context.reshape(B_context, C_context, HW).permute(0, 2, 1)
        context = self.context_proj(self.context_norm(context))
        context = context.repeat(B // context.shape[0], 1, 1)
        qk = torch.cat([x, context], dim=-1)
        qk = self.norm1(qk)
        v = self.norm1_v(x)

        q = (
            (self.q(qk) * self.scale)
            .reshape(B, H, W, self.num_heads, 16)
            .permute(0, 3, 1, 2, 4)
        )
        k = self.k(qk).reshape(B, H, W, self.num_heads, 16).permute(0, 3, 1, 2, 4)
        v = self.v(v).reshape(B, H, W, self.num_heads, 16).permute(0, 3, 1, 2, 4)

        attn = NATTENQKRPBFunction.apply(q, k, self.rpb)
        attn = attn.softmax(dim=-1)
        x = NATTENAVFunction.apply(attn, v)
        x = x.permute(0, 2, 3, 1, 4).reshape(B, HW, -1)
        x = self.proj(torch.cat([x, short_cut], dim=2))
        x = short_cut + self.drop_path(x)
        x = x + self.drop_path(self.ffn(self.norm2(x)))

        return x


from .twins import Block


class NATwins(nn.Module):
    def __init__(
        self,
        cost_latent_dim,
        dropout,
        droppath,
        vert_c_dim,
        encoder_latent_dim,
        attn_dim,
        expand_factor,
    ):
        super(NATwins, self).__init__()

        self.dim = dim = cost_latent_dim
        self.num_heads = num_heads = 8
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        embed_dim = dim
        mlp_ratio = 4
        ws = 7
        sr_ratio = 4
        dpr = droppath
        drop_rate = dropout
        attn_drop_rate = 0.0

        self.local_block = selfattentionlayer_nat(
            dropout=dropout,
            droppath=droppath,
            attn_dim=attn_dim,
            expand_factor=expand_factor,
            encoder_latent_dim=encoder_latent_dim,
            vert_c_dim=vert_c_dim,
            cost_latent_dim=cost_latent_dim,
        )
        self.global_block = Block(
            dim=embed_dim,
            num_heads=num_heads,
            mlp_ratio=mlp_ratio,
            drop=drop_rate,
            attn_drop=attn_drop_rate,
            drop_path=dpr,
            sr_ratio=sr_ratio,
            ws=1,
            with_rpe=True,
            vert_c_dim=vert_c_dim,
            encoder_latent_dim=encoder_latent_dim,
        )

    def forward(self, x, size, context):
        x = self.local_block(x, size, context)
        x = self.global_block(x, size, context)

        return x



""" Model creation / weight loading / state_dict helpers

Hacked together by / Copyright 2020 Ross Wightman
"""

import logging
import os
import math
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Callable, Optional, Tuple

import torch
import torch.nn as nn


from .features import FeatureListNet, FeatureDictNet, FeatureHookNet
from .hub import (
    has_hf_hub,
    download_cached_file,
    load_state_dict_from_hf,
    load_state_dict_from_url,
)
from .layers import Conv2dSame, Linear


_logger = logging.getLogger(__name__)


def load_state_dict(checkpoint_path, use_ema=False):
    if checkpoint_path and os.path.isfile(checkpoint_path):
        checkpoint = torch.load(checkpoint_path, map_location="cpu")
        state_dict_key = "state_dict"
        if isinstance(checkpoint, dict):
            if use_ema and "state_dict_ema" in checkpoint:
                state_dict_key = "state_dict_ema"
        if state_dict_key and state_dict_key in checkpoint:
            new_state_dict = OrderedDict()
            for k, v in checkpoint[state_dict_key].items():
                # strip `module.` prefix
                name = k[7:] if k.startswith("module") else k
                new_state_dict[name] = v
            state_dict = new_state_dict
        else:
            state_dict = checkpoint
        _logger.info(
            "Loaded {} from checkpoint '{}'".format(state_dict_key, checkpoint_path)
        )
        return state_dict
    else:
        _logger.error("No checkpoint found at '{}'".format(checkpoint_path))
        raise FileNotFoundError()


def load_checkpoint(model, checkpoint_path, use_ema=False, strict=True):
    if os.path.splitext(checkpoint_path)[-1].lower() in (".npz", ".npy"):
        # numpy checkpoint, try to load via model specific load_pretrained fn
        if hasattr(model, "load_pretrained"):
            model.load_pretrained(checkpoint_path)
        else:
            raise NotImplementedError("Model cannot load numpy checkpoint")
        return
    state_dict = load_state_dict(checkpoint_path, use_ema)
    model.load_state_dict(state_dict, strict=strict)


def resume_checkpoint(
    model, checkpoint_path, optimizer=None, loss_scaler=None, log_info=True
):
    resume_epoch = None
    if os.path.isfile(checkpoint_path):
        checkpoint = torch.load(checkpoint_path, map_location="cpu")
        if isinstance(checkpoint, dict) and "state_dict" in checkpoint:
            if log_info:
                _logger.info("Restoring model state from checkpoint...")
            new_state_dict = OrderedDict()
            for k, v in checkpoint["state_dict"].items():
                name = k[7:] if k.startswith("module") else k
                new_state_dict[name] = v
            model.load_state_dict(new_state_dict)

            if optimizer is not None and "optimizer" in checkpoint:
                if log_info:
                    _logger.info("Restoring optimizer state from checkpoint...")
                optimizer.load_state_dict(checkpoint["optimizer"])

            if loss_scaler is not None and loss_scaler.state_dict_key in checkpoint:
                if log_info:
                    _logger.info("Restoring AMP loss scaler state from checkpoint...")
                loss_scaler.load_state_dict(checkpoint[loss_scaler.state_dict_key])

            if "epoch" in checkpoint:
                resume_epoch = checkpoint["epoch"]
                if "version" in checkpoint and checkpoint["version"] > 1:
                    resume_epoch += 1  # start at the next epoch, old checkpoints incremented before save

            if log_info:
                _logger.info(
                    "Loaded checkpoint '{}' (epoch {})".format(
                        checkpoint_path, checkpoint["epoch"]
                    )
                )
        else:
            model.load_state_dict(checkpoint)
            if log_info:
                _logger.info("Loaded checkpoint '{}'".format(checkpoint_path))
        return resume_epoch
    else:
        _logger.error("No checkpoint found at '{}'".format(checkpoint_path))
        raise FileNotFoundError()


def load_custom_pretrained(
    model, default_cfg=None, load_fn=None, progress=False, check_hash=False
):
    r"""Loads a custom (read non .pth) weight file

    Downloads checkpoint file into cache-dir like torch.hub based loaders, but calls
    a passed in custom load fun, or the `load_pretrained` model member fn.

    If the object is already present in `model_dir`, it's deserialized and returned.
    The default value of `model_dir` is ``<hub_dir>/checkpoints`` where
    `hub_dir` is the directory returned by :func:`~torch.hub.get_dir`.

    Args:
        model: The instantiated model to load weights into
        default_cfg (dict): Default pretrained model cfg
        load_fn: An external stand alone fn that loads weights into provided model, otherwise a fn named
            'laod_pretrained' on the model will be called if it exists
        progress (bool, optional): whether or not to display a progress bar to stderr. Default: False
        check_hash(bool, optional): If True, the filename part of the URL should follow the naming convention
            ``filename-<sha256>.ext`` where ``<sha256>`` is the first eight or more
            digits of the SHA256 hash of the contents of the file. The hash is used to
            ensure unique names and to verify the contents of the file. Default: False
    """
    default_cfg = default_cfg or getattr(model, "default_cfg", None) or {}
    pretrained_url = default_cfg.get("url", None)
    if not pretrained_url:
        _logger.warning(
            "No pretrained weights exist for this model. Using random initialization."
        )
        return
    cached_file = download_cached_file(
        default_cfg["url"], check_hash=check_hash, progress=progress
    )

    if load_fn is not None:
        load_fn(model, cached_file)
    elif hasattr(model, "load_pretrained"):
        model.load_pretrained(cached_file)
    else:
        _logger.warning(
            "Valid function to load pretrained weights is not available, using random initialization."
        )


def adapt_input_conv(in_chans, conv_weight):
    conv_type = conv_weight.dtype
    conv_weight = (
        conv_weight.float()
    )  # Some weights are in torch.half, ensure it's float for sum on CPU
    O, I, J, K = conv_weight.shape
    if in_chans == 1:
        if I > 3:
            assert conv_weight.shape[1] % 3 == 0
            # For models with space2depth stems
            conv_weight = conv_weight.reshape(O, I // 3, 3, J, K)
            conv_weight = conv_weight.sum(dim=2, keepdim=False)
        else:
            conv_weight = conv_weight.sum(dim=1, keepdim=True)
    elif in_chans != 3:
        if I != 3:
            raise NotImplementedError("Weight format not supported by conversion.")
        else:
            # NOTE this strategy should be better than random init, but there could be other combinations of
            # the original RGB input layer weights that'd work better for specific cases.
            repeat = int(math.ceil(in_chans / 3))
            conv_weight = conv_weight.repeat(1, repeat, 1, 1)[:, :in_chans, :, :]
            conv_weight *= 3 / float(in_chans)
    conv_weight = conv_weight.to(conv_type)
    return conv_weight


def load_pretrained(
    model,
    default_cfg=None,
    num_classes=1000,
    in_chans=3,
    filter_fn=None,
    strict=True,
    progress=False,
):
    """Load pretrained checkpoint

    Args:
        model (nn.Module) : PyTorch model module
        default_cfg (Optional[Dict]): default configuration for pretrained weights / target dataset
        num_classes (int): num_classes for model
        in_chans (int): in_chans for model
        filter_fn (Optional[Callable]): state_dict filter fn for load (takes state_dict, model as args)
        strict (bool): strict load of checkpoint
        progress (bool): enable progress bar for weight download

    """
    default_cfg = default_cfg or getattr(model, "default_cfg", None) or {}
    pretrained_url = default_cfg.get("url", None)
    hf_hub_id = default_cfg.get("hf_hub", None)
    if not pretrained_url and not hf_hub_id:
        _logger.warning(
            "No pretrained weights exist for this model. Using random initialization."
        )
        return
    if hf_hub_id and has_hf_hub(necessary=not pretrained_url):
        _logger.info(f"Loading pretrained weights from Hugging Face hub ({hf_hub_id})")
        state_dict = load_state_dict_from_hf(hf_hub_id)
    else:
        _logger.info(f"Loading pretrained weights from url ({pretrained_url})")
        state_dict = load_state_dict_from_url(
            pretrained_url, progress=progress, map_location="cpu"
        )
    if filter_fn is not None:
        # for backwards compat with filter fn that take one arg, try one first, the two
        try:
            state_dict = filter_fn(state_dict)
        except TypeError:
            state_dict = filter_fn(state_dict, model)

    input_convs = default_cfg.get("first_conv", None)
    if input_convs is not None and in_chans != 3:
        if isinstance(input_convs, str):
            input_convs = (input_convs,)
        for input_conv_name in input_convs:
            weight_name = input_conv_name + ".weight"
            try:
                state_dict[weight_name] = adapt_input_conv(
                    in_chans, state_dict[weight_name]
                )
                _logger.info(
                    f"Converted input conv {input_conv_name} pretrained weights from 3 to {in_chans} channel(s)"
                )
            except NotImplementedError as e:
                del state_dict[weight_name]
                strict = False
                _logger.warning(
                    f"Unable to convert pretrained {input_conv_name} weights, using random init for this layer."
                )

    classifiers = default_cfg.get("classifier", None)
    label_offset = default_cfg.get("label_offset", 0)
    if classifiers is not None:
        if isinstance(classifiers, str):
            classifiers = (classifiers,)
        if num_classes != default_cfg["num_classes"]:
            for classifier_name in classifiers:
                # completely discard fully connected if model num_classes doesn't match pretrained weights
                del state_dict[classifier_name + ".weight"]
                del state_dict[classifier_name + ".bias"]
            strict = False
        elif label_offset > 0:
            for classifier_name in classifiers:
                # special case for pretrained weights with an extra background class in pretrained weights
                classifier_weight = state_dict[classifier_name + ".weight"]
                state_dict[classifier_name + ".weight"] = classifier_weight[
                    label_offset:
                ]
                classifier_bias = state_dict[classifier_name + ".bias"]
                state_dict[classifier_name + ".bias"] = classifier_bias[label_offset:]

    model.load_state_dict(state_dict, strict=strict)


def extract_layer(model, layer):
    layer = layer.split(".")
    module = model
    if hasattr(model, "module") and layer[0] != "module":
        module = model.module
    if not hasattr(model, "module") and layer[0] == "module":
        layer = layer[1:]
    for l in layer:
        if hasattr(module, l):
            if not l.isdigit():
                module = getattr(module, l)
            else:
                module = module[int(l)]
        else:
            return module
    return module


def set_layer(model, layer, val):
    layer = layer.split(".")
    module = model
    if hasattr(model, "module") and layer[0] != "module":
        module = model.module
    lst_index = 0
    module2 = module
    for l in layer:
        if hasattr(module2, l):
            if not l.isdigit():
                module2 = getattr(module2, l)
            else:
                module2 = module2[int(l)]
            lst_index += 1
    lst_index -= 1
    for l in layer[:lst_index]:
        if not l.isdigit():
            module = getattr(module, l)
        else:
            module = module[int(l)]
    l = layer[lst_index]
    setattr(module, l, val)


def adapt_model_from_string(parent_module, model_string):
    separator = "***"
    state_dict = {}
    lst_shape = model_string.split(separator)
    for k in lst_shape:
        k = k.split(":")
        key = k[0]
        shape = k[1][1:-1].split(",")
        if shape[0] != "":
            state_dict[key] = [int(i) for i in shape]

    new_module = deepcopy(parent_module)
    for n, m in parent_module.named_modules():
        old_module = extract_layer(parent_module, n)
        if isinstance(old_module, nn.Conv2d) or isinstance(old_module, Conv2dSame):
            if isinstance(old_module, Conv2dSame):
                conv = Conv2dSame
            else:
                conv = nn.Conv2d
            s = state_dict[n + ".weight"]
            in_channels = s[1]
            out_channels = s[0]
            g = 1
            if old_module.groups > 1:
                in_channels = out_channels
                g = in_channels
            new_conv = conv(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=old_module.kernel_size,
                bias=old_module.bias is not None,
                padding=old_module.padding,
                dilation=old_module.dilation,
                groups=g,
                stride=old_module.stride,
            )
            set_layer(new_module, n, new_conv)
        if isinstance(old_module, nn.BatchNorm2d):
            new_bn = nn.BatchNorm2d(
                num_features=state_dict[n + ".weight"][0],
                eps=old_module.eps,
                momentum=old_module.momentum,
                affine=old_module.affine,
                track_running_stats=True,
            )
            set_layer(new_module, n, new_bn)
        if isinstance(old_module, nn.Linear):
            # FIXME extra checks to ensure this is actually the FC classifier layer and not a diff Linear layer?
            num_features = state_dict[n + ".weight"][1]
            new_fc = Linear(
                in_features=num_features,
                out_features=old_module.out_features,
                bias=old_module.bias is not None,
            )
            set_layer(new_module, n, new_fc)
            if hasattr(new_module, "num_features"):
                new_module.num_features = num_features
    new_module.eval()
    parent_module.eval()

    return new_module


def adapt_model_from_file(parent_module, model_variant):
    adapt_file = os.path.join(
        os.path.dirname(__file__), "pruned", model_variant + ".txt"
    )
    with open(adapt_file, "r") as f:
        return adapt_model_from_string(parent_module, f.read().strip())


def default_cfg_for_features(default_cfg):
    default_cfg = deepcopy(default_cfg)
    # remove default pretrained cfg fields that don't have much relevance for feature backbone
    to_remove = (
        "num_classes",
        "crop_pct",
        "classifier",
        "global_pool",
    )  # add default final pool size?
    for tr in to_remove:
        default_cfg.pop(tr, None)
    return default_cfg


def overlay_external_default_cfg(default_cfg, kwargs):
    """Overlay 'external_default_cfg' in kwargs on top of default_cfg arg."""
    external_default_cfg = kwargs.pop("external_default_cfg", None)
    if external_default_cfg:
        default_cfg.pop("url", None)  # url should come from external cfg
        default_cfg.pop("hf_hub", None)  # hf hub id should come from external cfg
        default_cfg.update(external_default_cfg)


def set_default_kwargs(kwargs, names, default_cfg):
    for n in names:
        # for legacy reasons, model __init__args uses img_size + in_chans as separate args while
        # default_cfg has one input_size=(C, H ,W) entry
        if n == "img_size":
            input_size = default_cfg.get("input_size", None)
            if input_size is not None:
                assert len(input_size) == 3
                kwargs.setdefault(n, input_size[-2:])
        elif n == "in_chans":
            input_size = default_cfg.get("input_size", None)
            if input_size is not None:
                assert len(input_size) == 3
                kwargs.setdefault(n, input_size[0])
        else:
            default_val = default_cfg.get(n, None)
            if default_val is not None:
                kwargs.setdefault(n, default_cfg[n])


def filter_kwargs(kwargs, names):
    if not kwargs or not names:
        return
    for n in names:
        kwargs.pop(n, None)


def update_default_cfg_and_kwargs(default_cfg, kwargs, kwargs_filter):
    """Update the default_cfg and kwargs before passing to model

    FIXME this sequence of overlay default_cfg, set default kwargs, filter kwargs
    could/should be replaced by an improved configuration mechanism

    Args:
        default_cfg: input default_cfg (updated in-place)
        kwargs: keyword args passed to model build fn (updated in-place)
        kwargs_filter: keyword arg keys that must be removed before model __init__
    """
    # Overlay default cfg values from `external_default_cfg` if it exists in kwargs
    overlay_external_default_cfg(default_cfg, kwargs)
    # Set model __init__ args that can be determined by default_cfg (if not already passed as kwargs)
    default_kwarg_names = ("num_classes", "global_pool", "in_chans")
    if default_cfg.get("fixed_input_size", False):
        # if fixed_input_size exists and is True, model takes an img_size arg that fixes its input size
        default_kwarg_names += ("img_size",)
    set_default_kwargs(kwargs, names=default_kwarg_names, default_cfg=default_cfg)
    # Filter keyword args for task specific model variants (some 'features only' models, etc.)
    filter_kwargs(kwargs, names=kwargs_filter)


def build_model_with_cfg(
    model_cls: Callable,
    variant: str,
    pretrained: bool,
    default_cfg: dict,
    model_cfg: Optional[Any] = None,
    feature_cfg: Optional[dict] = None,
    pretrained_strict: bool = True,
    pretrained_filter_fn: Optional[Callable] = None,
    pretrained_custom_load: bool = False,
    kwargs_filter: Optional[Tuple[str]] = None,
    **kwargs,
):
    """Build model with specified default_cfg and optional model_cfg

    This helper fn aids in the construction of a model including:
      * handling default_cfg and associated pretained weight loading
      * passing through optional model_cfg for models with config based arch spec
      * features_only model adaptation
      * pruning config / model adaptation

    Args:
        model_cls (nn.Module): model class
        variant (str): model variant name
        pretrained (bool): load pretrained weights
        default_cfg (dict): model's default pretrained/task config
        model_cfg (Optional[Dict]): model's architecture config
        feature_cfg (Optional[Dict]: feature extraction adapter config
        pretrained_strict (bool): load pretrained weights strictly
        pretrained_filter_fn (Optional[Callable]): filter callable for pretrained weights
        pretrained_custom_load (bool): use custom load fn, to load numpy or other non PyTorch weights
        kwargs_filter (Optional[Tuple]): kwargs to filter before passing to model
        **kwargs: model args passed through to model __init__
    """
    pruned = kwargs.pop("pruned", False)
    features = False
    feature_cfg = feature_cfg or {}
    default_cfg = deepcopy(default_cfg) if default_cfg else {}
    update_default_cfg_and_kwargs(default_cfg, kwargs, kwargs_filter)
    default_cfg.setdefault("architecture", variant)

    # Setup for feature extraction wrapper done at end of this fn
    if kwargs.pop("features_only", False):
        features = True
        feature_cfg.setdefault("out_indices", (0, 1, 2, 3, 4))
        if "out_indices" in kwargs:
            feature_cfg["out_indices"] = kwargs.pop("out_indices")

    # Build the model
    model = (
        model_cls(**kwargs) if model_cfg is None else model_cls(cfg=model_cfg, **kwargs)
    )
    model.default_cfg = default_cfg

    if pruned:
        model = adapt_model_from_file(model, variant)

    # For classification models, check class attr, then kwargs, then default to 1k, otherwise 0 for feats
    num_classes_pretrained = (
        0
        if features
        else getattr(model, "num_classes", kwargs.get("num_classes", 1000))
    )
    if pretrained:
        if pretrained_custom_load:
            load_custom_pretrained(model)
        else:
            load_pretrained(
                model,
                num_classes=num_classes_pretrained,
                in_chans=kwargs.get("in_chans", 3),
                filter_fn=pretrained_filter_fn,
                strict=pretrained_strict,
            )

    # Wrap the model in a feature extraction module if enabled
    if features:
        feature_cls = FeatureListNet
        if "feature_cls" in feature_cfg:
            feature_cls = feature_cfg.pop("feature_cls")
            if isinstance(feature_cls, str):
                feature_cls = feature_cls.lower()
                if "hook" in feature_cls:
                    feature_cls = FeatureHookNet
                else:
                    assert False, f"Unknown feature class {feature_cls}"
        model = feature_cls(model, **feature_cfg)
        model.default_cfg = default_cfg_for_features(
            default_cfg
        )  # add back default_cfg

    return model


def model_parameters(model, exclude_head=False):
    if exclude_head:
        # FIXME this a bit of a quick and dirty hack to skip classifier head params based on ordering
        return [p for p in model.parameters()][:-2]
    else:
        return model.parameters()


def named_apply(
    fn: Callable, module: nn.Module, name="", depth_first=True, include_root=False
) -> nn.Module:
    if not depth_first and include_root:
        fn(module=module, name=name)
    for child_name, child_module in module.named_children():
        child_name = ".".join((name, child_name)) if name else child_name
        named_apply(
            fn=fn,
            module=child_module,
            name=child_name,
            depth_first=depth_first,
            include_root=True,
        )
    if depth_first and include_root:
        fn(module=module, name=name)
    return module


def named_modules(module: nn.Module, name="", depth_first=True, include_root=False):
    if not depth_first and include_root:
        yield name, module
    for child_name, child_module in module.named_children():
        child_name = ".".join((name, child_name)) if name else child_name
        yield from named_modules(
            module=child_module,
            name=child_name,
            depth_first=depth_first,
            include_root=True,
        )
    if depth_first and include_root:
        yield name, module



import torch.nn as nn


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution without padding"""
    return nn.Conv2d(
        in_planes, out_planes, kernel_size=1, stride=stride, padding=0, bias=False
    )


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(
        in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False
    )


class BasicBlock(nn.Module):
    def __init__(self, in_planes, planes, stride=1):
        super().__init__()
        self.conv1 = conv3x3(in_planes, planes, stride)
        self.conv2 = conv3x3(planes, planes)
        self.bn1 = nn.BatchNorm2d(planes)
        self.bn2 = nn.BatchNorm2d(planes)

        self.relu = nn.ReLU(inplace=True)

        if stride == 1:
            self.downsample = None
        else:
            self.downsample = nn.Sequential(
                conv1x1(in_planes, planes, stride=stride), nn.BatchNorm2d(planes)
            )

    def forward(self, x):
        y = x
        y = self.relu(self.bn1(self.conv1(y)))
        y = self.bn2(self.conv2(y))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x + y)


class ResNetFPN_8_2(nn.Module):
    """
    ResNet+FPN, output resolution are 1/8 and 1/2.
    Each block has 2 layers.
    """

    def __init__(self):
        super().__init__()
        # Config
        block = BasicBlock
        initial_dim = 128
        block_dims = [128, 196, 256]

        # Class Variable
        self.in_planes = initial_dim
        self.initial_dim = initial_dim

        # Networks
        self.conv1 = nn.Conv2d(
            3, initial_dim, kernel_size=7, stride=2, padding=3, bias=False
        )
        self.bn1 = nn.BatchNorm2d(initial_dim)

        self.relu = nn.ReLU(inplace=True)

        self.layer1 = self._make_layer(block, block_dims[0], stride=1)  # 1/2
        self.layer2 = self._make_layer(block, block_dims[1], stride=2)  # 1/4
        self.layer3 = self._make_layer(block, block_dims[2], stride=2)  # 1/8

        # 3. FPN upsample
        self.layer3_outconv = conv1x1(block_dims[2], block_dims[2])

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def _make_layer(self, block, dim, stride=1):
        layer1 = block(self.in_planes, dim, stride=stride)
        layer2 = block(dim, dim, stride=1)
        layers = (layer1, layer2)

        self.in_planes = dim
        return nn.Sequential(*layers)

    def forward(self, x):
        # ResNet Backbone
        x0 = self.relu(self.bn1(self.conv1(x)))
        x1 = self.layer1(x0)  # 1/2
        x2 = self.layer2(x1)  # 1/4
        x3 = self.layer3(x2)  # 1/8

        x3_out = self.layer3_outconv(x3)
        return x3_out



import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from ptlflow.utils.utils import forward_interpolate_batch
from .update import BasicUpdateBlock, SmallUpdateBlock
from .extractor import BasicEncoderQuarter
from .utils import upflow4
from .path_match import PathMatch
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += i_weight * (valid * i_loss).mean()

        return flow_loss


class DIP(BaseModel):
    pretrained_checkpoints = {
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/dip-kitti-b0b678b4.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/dip-sintel-7abeb652.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/dip-things-688d52a0.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        dropout: float = 0.0,
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 20,
        max_offset: int = 256,
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=16, loss_fn=SequenceLoss(gamma, max_flow), **kwargs
        )

        self.corr_levels = corr_levels
        self.corr_radius = corr_radius
        self.dropout = dropout
        self.gamma = gamma
        self.max_flow = max_flow
        self.iters = iters
        self.max_offset = max_offset

        self.hidden_dim = 128
        self.context_dim = 128

        self.dropout = 0

        # feature network, and update block
        self.fnet = BasicEncoderQuarter(
            output_dim=256, norm_fn="instance", dropout=self.dropout
        )

        self.update_block_s = SmallUpdateBlock(hidden_dim=self.hidden_dim)
        self.update_block = BasicUpdateBlock(hidden_dim=self.hidden_dim)

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def random_init_flow(self, fmap, max_offset, test_mode=False):
        N, C, H, W = fmap.shape
        if test_mode:
            init_seed = 20
            torch.manual_seed(init_seed)
            torch.cuda.manual_seed(init_seed)
        flow = (torch.rand(N, 2, H, W) - 0.5) * 2
        flow = flow.to(dtype=fmap.dtype, device=fmap.device) * max_offset
        return flow

    def upsample_flow(self, flow, mask, rate=4):
        """Upsample flow field [H/rate, W/rate, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, rate, rate, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(rate * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, rate * H, rate * W)

    def build_pyramid(self, fmap1, fmap2, cnet, max_layers=5, min_width=40):
        py_fmap1 = []
        py_fmap2 = []
        py_cnet = []
        py_fmap1.append(fmap1)
        py_fmap2.append(fmap2)
        py_cnet.append(cnet)

        curr_fmap1 = fmap1
        curr_fmap2 = fmap2
        curr_cnet = cnet
        for i in range(max_layers - 1):
            if (curr_fmap1.shape[2] < min_width) and (curr_fmap1.shape[3] < min_width):
                break
            down_scale = 2 ** (i + 1)
            curr_fmap1 = F.avg_pool2d(curr_fmap1, 2, stride=2)
            curr_fmap2 = F.avg_pool2d(curr_fmap2, 2, stride=2)
            curr_cnet = F.avg_pool2d(curr_cnet, 2, stride=2)
            py_fmap1.append(curr_fmap1)
            py_fmap2.append(curr_fmap2)
            py_cnet.append(curr_cnet)

        return py_fmap1, py_fmap2, py_cnet

    def upflow(self, flow, targetMap, mode="bilinear"):
        """Upsample flow"""
        new_size = (targetMap.shape[2], targetMap.shape[3])
        factor = 1.0 * targetMap.shape[2] / flow.shape[2]
        return factor * F.interpolate(
            flow, size=new_size, mode=mode, align_corners=True
        )

    def inference(self, image1, image2, iters=6, max_layers=3, init_flow=None):
        max_layers = 3
        max_offset = 256
        auto_layer = 3
        if init_flow is not None:
            mag = torch.norm(init_flow, dim=1)
            max_offset = torch.max(mag)
            auto_layer = max_offset / 32
            auto_layer = int(auto_layer.ceil().cpu().numpy())
            if auto_layer < max_layers:
                max_layers = auto_layer

        # run the feature network
        fmap1, fmap2 = self.fnet([image1, image2])
        new_size = (image1.shape[2], image1.shape[3])

        # build layers
        min_width = 40
        py_fmap1, py_fmap2, py_cnet = self.build_pyramid(
            fmap1, fmap2, fmap1, max_layers=max_layers, min_width=min_width
        )
        n_levels = len(py_fmap1)

        # build layers
        s_fmap1 = py_fmap1[n_levels - 1]
        # init
        if init_flow is not None:
            new_size = (s_fmap1.shape[2], s_fmap1.shape[3])
            scale = s_fmap1.shape[2] / (init_flow.shape[2] * 1.0)
            s_flow = scale * F.interpolate(
                init_flow, size=new_size, mode="bilinear", align_corners=True
            )
            initail_flow_max = 2 ** (auto_layer + 1) * 1.0
            noise = self.random_init_flow(
                s_fmap1, max_offset=16, test_mode=(not self.training)
            )
            s_flow = s_flow + noise
        else:
            scale = 2 ** (n_levels - 1 + 2) * 1.0
            s_flow = self.random_init_flow(
                s_fmap1,
                max_offset=self.max_offset / scale,
                test_mode=(not self.training),
            )

        up_mask = None
        for i in range(n_levels):
            curr_fmap1 = py_fmap1[n_levels - i - 1]
            curr_fmap2 = py_fmap2[n_levels - i - 1]
            curr_cnet = py_cnet[n_levels - i - 1]
            patch_fn = PathMatch(curr_fmap1, curr_fmap2)
            net, inp = torch.split(
                curr_cnet, [self.hidden_dim, self.context_dim], dim=1
            )
            net = torch.tanh(net)
            inp = torch.relu(inp)
            if i > 0:
                s_flow = self.upflow(flow_up, curr_fmap1)
                noise = self.random_init_flow(
                    curr_fmap1, max_offset=4, test_mode=(not self.training)
                )
                s_flow = s_flow + noise

            for itr in range(iters):
                s_flow = s_flow.detach()
                out_corrs = patch_fn(s_flow, is_search=False)

                net, up_mask, delta_flow = self.update_block_s(
                    net, inp, out_corrs, s_flow
                )

                s_flow = s_flow + delta_flow

                s_flow = s_flow.detach()
                out_corrs = patch_fn(s_flow, is_search=True)
                net, up_mask, delta_flow = self.update_block(
                    net, inp, out_corrs, s_flow
                )

                s_flow = s_flow + delta_flow
                flow_up = self.upsample_flow(s_flow, up_mask, rate=4)

        return flow_up

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=False,
            resize_mode="pad",
            pad_mode="constant",
            pad_two_side=True,
            pad_value=-1,
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        init_flow = None
        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            init_flow = forward_interpolate_batch(inputs["prev_preds"]["flow_small"])

        if not self.training and init_flow is not None:
            flow_up = self.inference(
                image1, image2, iters=self.iters, init_flow=init_flow
            )
            flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
            return {"flows": flow_up[:, None]}

        # run the feature network
        fmap1, fmap2 = self.fnet([image1, image2])

        # run the context network
        net, inp = torch.split(fmap1, [self.hidden_dim, self.context_dim], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)

        # 1/4 -> 1/16
        # feature
        s_fmap1 = F.avg_pool2d(fmap1, 4, stride=4)
        s_fmap2 = F.avg_pool2d(fmap2, 4, stride=4)
        # context(left)
        s_net = F.avg_pool2d(net, 4, stride=4)
        s_inp = F.avg_pool2d(inp, 4, stride=4)

        # 1/16
        s_patch_fn = PathMatch(s_fmap1, s_fmap2)

        # init flow
        s_flow = None
        s_flow = self.random_init_flow(
            s_fmap1,
            max_offset=self.max_offset // 16,
            test_mode=(not self.training),
        )

        # small initial: 1/16
        flow = None
        flow_up = None
        flow_predictions = []
        for itr in range(self.iters):
            # --------------- update1 ---------------
            s_flow = s_flow.detach()
            out_corrs = s_patch_fn(s_flow, is_search=False)

            s_net, up_mask, delta_flow = self.update_block_s(
                s_net, s_inp, out_corrs, s_flow
            )

            s_flow = s_flow + delta_flow
            flow = self.upsample_flow(s_flow, up_mask, rate=4)
            flow_up = upflow4(flow)
            flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
            flow_predictions.append(flow_up)

            # --------------- update2 ---------------
            s_flow = s_flow.detach()
            out_corrs = s_patch_fn(s_flow, is_search=True)

            s_net, up_mask, delta_flow = self.update_block(
                s_net, s_inp, out_corrs, s_flow
            )

            s_flow = s_flow + delta_flow
            flow = self.upsample_flow(s_flow, up_mask, rate=4)
            flow_up = upflow4(flow)
            flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
            flow_predictions.append(flow_up)

        patch_fn = PathMatch(fmap1, fmap2)
        # large refine: 1/4
        for itr in range(self.iters):
            # --------------- update1 ---------------
            flow = flow.detach()
            out_corrs = patch_fn(flow, is_search=False)
            net, up_mask, delta_flow = self.update_block_s(net, inp, out_corrs, flow)

            flow = flow + delta_flow
            flow_up = self.upsample_flow(flow, up_mask, rate=4)
            flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
            flow_predictions.append(flow_up)

            # --------------- update2 ---------------
            flow = flow.detach()
            out_corrs = patch_fn(flow, is_search=True)
            net, up_mask, delta_flow = self.update_block(net, inp, out_corrs, flow)

            flow = flow + delta_flow
            flow_up = self.upsample_flow(flow, up_mask, rate=4)
            flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
            flow_predictions.append(flow_up)

        if self.training:
            outputs = {"flows": flow_up[:, None], "flow_preds": flow_predictions}
        else:
            outputs = {"flows": flow_up[:, None], "flow_small": flow}

        return outputs


@register_model
@trainable
class dip(DIP):
    pass



import torch
import torch.nn as nn
import torch.nn.functional as F


class FlowHead(nn.Module):
    def __init__(self, input_dim=128, hidden_dim=256):
        super(FlowHead, self).__init__()
        self.conv1 = nn.Conv2d(input_dim, hidden_dim, 3, padding=1)
        self.conv2 = nn.Conv2d(hidden_dim, 2, 3, padding=1)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        return self.conv2(self.relu(self.conv1(x)))


class ConvGRU(nn.Module):
    def __init__(self, hidden_dim=128, input_dim=192 + 128):
        super(ConvGRU, self).__init__()
        self.convz = nn.Conv2d(hidden_dim + input_dim, hidden_dim, 3, padding=1)
        self.convr = nn.Conv2d(hidden_dim + input_dim, hidden_dim, 3, padding=1)
        self.convq = nn.Conv2d(hidden_dim + input_dim, hidden_dim, 3, padding=1)

    def forward(self, h, x):
        hx = torch.cat([h, x], dim=1)

        z = torch.sigmoid(self.convz(hx))
        r = torch.sigmoid(self.convr(hx))
        q = torch.tanh(self.convq(torch.cat([r * h, x], dim=1)))

        h = (1 - z) * h + z * q
        return h


class SepConvGRU(nn.Module):
    def __init__(
        self, hidden_dim=128, input_dim=256 + 128
    ):  # input dim=256 + 128, hdim=64
        super(SepConvGRU, self).__init__()
        self.convz1 = nn.Conv2d(input_dim, hidden_dim, (1, 5), padding=(0, 2))
        self.convr1 = nn.Conv2d(input_dim, hidden_dim, (1, 5), padding=(0, 2))
        self.convq1 = nn.Conv2d(input_dim, hidden_dim, (1, 5), padding=(0, 2))

        self.convz2 = nn.Conv2d(input_dim, hidden_dim, (5, 1), padding=(2, 0))
        self.convr2 = nn.Conv2d(input_dim, hidden_dim, (5, 1), padding=(2, 0))
        self.convq2 = nn.Conv2d(input_dim, hidden_dim, (5, 1), padding=(2, 0))

    # h:net:128, x = 128 + 128 = 256
    def forward(self, h, x):
        # horizontal
        hx = torch.cat([h, x], dim=1)
        z = torch.sigmoid(self.convz1(hx))
        r = torch.sigmoid(self.convr1(hx))
        q = torch.tanh(self.convq1(torch.cat([r * h, x], dim=1)))
        h = (1 - z) * h + z * q

        # vertical
        hx = torch.cat([h, x], dim=1)
        z = torch.sigmoid(self.convz2(hx))
        r = torch.sigmoid(self.convr2(hx))
        q = torch.tanh(self.convq2(torch.cat([r * h, x], dim=1)))
        h = (1 - z) * h + z * q

        return h


class BasicMotionEncoder(nn.Module):
    def __init__(self, correlation_depth, stack_coords=False):
        super(BasicMotionEncoder, self).__init__()
        self.convc1 = nn.Conv2d(correlation_depth, 256, 1, padding=0)
        self.stack_coords = stack_coords
        # self.convc1 = nn.Conv2d(324, 256, 1, padding=0)
        self.convc2 = nn.Conv2d(256, 192, 3, padding=1)
        self.convf1 = nn.Conv2d(2, 128, 7, padding=3)
        self.convf2 = nn.Conv2d(128, 64, 3, padding=1)
        if not stack_coords:
            self.conv = nn.Conv2d(64 + 192, 128 - 2, 3, padding=1)

        else:
            self.conv_coords_x_1 = nn.Conv2d(correlation_depth, 64, 3, padding=1)
            self.conv_coords_x_2 = nn.Conv2d(64, 64, 3, padding=1)

            self.conv_coords_y_1 = nn.Conv2d(correlation_depth, 64, 3, padding=1)
            self.conv_coords_y_2 = nn.Conv2d(64, 64, 3, padding=1)

            self.conv_corr_coords = nn.Conv2d(192 + 128, 256, 3, padding=1)
            self.conv_corr_coords_flow_1 = nn.Conv2d(256 + 64, 256, 3, padding=1)
            self.conv_corr_coords_flow_2 = nn.Conv2d(256, 128 - 2, 1, padding=0)

    def forward(self, flow, corr, coords_x=None, coords_y=None):
        cor = F.relu(self.convc1(corr))
        cor = F.relu(self.convc2(cor))
        flo = F.relu(self.convf1(flow))
        flo = F.relu(self.convf2(flo))
        if not self.stack_coords:
            cor_flo = torch.cat([cor, flo], dim=1)
            out = F.relu(self.conv(cor_flo))
        else:
            conved_coords_x_1 = F.relu(self.conv_coords_x_1(coords_x))
            conved2_coords_x = F.relu(self.conv_coords_x_2(conved_coords_x_1))

            conved_coords_y_1 = F.relu(self.conv_coords_y_1(coords_y))
            conved2_coords_y = F.relu(self.conv_coords_y_2(conved_coords_y_1))

            conved_corr_coords = F.relu(
                self.conv_corr_coords(
                    torch.cat([conved2_coords_x, conved2_coords_y, cor], dim=1)
                )
            )
            conved_corr_coords_flow_1 = F.relu(
                self.conv_corr_coords_flow_1(
                    torch.cat([conved_corr_coords, flo], dim=1)
                )
            )
            out = F.relu(self.conv_corr_coords_flow_2(conved_corr_coords_flow_1))
        return torch.cat([out, flow], dim=1)


class BasicUpdateBlock(nn.Module):
    def __init__(
        self,
        correlation_depth,
        stack_coords=False,
        hidden_dim=128,
        input_dim=256,
        scale=8,
    ):
        super(BasicUpdateBlock, self).__init__()
        self.encoder = BasicMotionEncoder(correlation_depth, stack_coords=stack_coords)
        self.gru = SepConvGRU(hidden_dim=hidden_dim, input_dim=input_dim + hidden_dim)
        self.flow_head = FlowHead(hidden_dim, hidden_dim=256)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, scale * scale * 9, 1, padding=0),
        )

    def forward(self, net, inp, corr, flow, coords_x=None, coords_y=None):
        # net: 128
        # inp depth: 128
        motion_features = self.encoder(
            flow, corr, coords_x, coords_y
        )  # motion feature depth: 128
        inp = torch.cat([inp, motion_features], dim=1)

        net = self.gru(net, inp)  # output:net.depth:128

        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow



import torch
from torch.optim.lr_scheduler import OneCycleLR


def fetch_optimizer(
    model,
    optimizer_name,
    scheduler_name,
    canonical_lr,
    twins_lr_factor,
    adam_decay,
    adamw_decay,
    epsilon,
    num_steps,
    anneal_strategy,
):
    """Create the optimizer and learning rate scheduler"""
    optimizer = build_optimizer(
        model,
        optimizer_name,
        canonical_lr,
        twins_lr_factor,
        adam_decay,
        adamw_decay,
        epsilon,
    )
    scheduler = build_scheduler(
        scheduler_name,
        canonical_lr,
        twins_lr_factor,
        num_steps,
        anneal_strategy,
        optimizer,
    )

    return optimizer, scheduler


def build_optimizer(
    model,
    optimizer_name,
    canonical_lr,
    twins_lr_factor,
    adam_decay,
    adamw_decay,
    epsilon,
):
    name = optimizer_name
    lr = canonical_lr

    if name == "adam":
        return torch.optim.Adam(
            model.parameters(),
            lr=lr,
            weight_decay=adam_decay,
            eps=epsilon,
        )
    elif name == "adamw":
        if twins_lr_factor is not None:
            factor = twins_lr_factor
            print("[Decrease lr of pre-trained model by factor {}]".format(factor))
            param_dicts = [
                {
                    "params": [
                        p
                        for n, p in model.named_parameters()
                        if "feat_encoder" not in n
                        and "context_encoder" not in n
                        and p.requires_grad
                    ]
                },
                {
                    "params": [
                        p
                        for n, p in model.named_parameters()
                        if ("feat_encoder" in n or "context_encoder" in n)
                        and p.requires_grad
                    ],
                    "lr": lr * factor,
                },
            ]
            full = [n for n, _ in model.named_parameters()]
            return torch.optim.AdamW(
                param_dicts, lr=lr, weight_decay=adamw_decay, eps=epsilon
            )
        else:
            return torch.optim.AdamW(
                model.parameters(),
                lr=lr,
                weight_decay=adamw_decay,
                eps=epsilon,
            )
    else:
        raise ValueError(f"TRAINER.OPTIMIZER = {name} is not a valid optimizer!")


def build_scheduler(
    scheduler_name, canonical_lr, twins_lr_factor, num_steps, anneal_strategy, optimizer
):
    """
    Returns:
        scheduler (dict):{
            'scheduler': lr_scheduler,
            'interval': 'step',  # or 'epoch'
        }
    """
    name = scheduler_name
    lr = canonical_lr

    if name == "OneCycleLR":
        if twins_lr_factor is not None:
            factor = twins_lr_factor
            scheduler = OneCycleLR(
                optimizer,
                [lr, lr * factor],
                num_steps + 100,
                pct_start=0.05,
                cycle_momentum=False,
                anneal_strategy=anneal_strategy,
            )
        else:
            scheduler = OneCycleLR(
                optimizer,
                lr,
                num_steps + 100,
                pct_start=0.05,
                cycle_momentum=False,
                anneal_strategy=anneal_strategy,
            )
    else:
        raise NotImplementedError()

    return scheduler



""" Model creation / weight loading / state_dict helpers

Hacked together by / Copyright 2020 Ross Wightman
"""

from itertools import chain

import torch
from torch.utils.checkpoint import checkpoint


def checkpoint_seq(
    functions, x, every=1, flatten=False, skip_last=False, preserve_rng_state=True
):
    r"""A helper function for checkpointing sequential models.

    Sequential models execute a list of modules/functions in order
    (sequentially). Therefore, we can divide such a sequence into segments
    and checkpoint each segment. All segments except run in :func:`torch.no_grad`
    manner, i.e., not storing the intermediate activations. The inputs of each
    checkpointed segment will be saved for re-running the segment in the backward pass.

    See :func:`~torch.utils.checkpoint.checkpoint` on how checkpointing works.

    .. warning::
        Checkpointing currently only supports :func:`torch.autograd.backward`
        and only if its `inputs` argument is not passed. :func:`torch.autograd.grad`
        is not supported.

    .. warning:
        At least one of the inputs needs to have :code:`requires_grad=True` if
        grads are needed for model inputs, otherwise the checkpointed part of the
        model won't have gradients.

    Args:
        functions: A :class:`torch.nn.Sequential` or the list of modules or functions to run sequentially.
        x: A Tensor that is input to :attr:`functions`
        every: checkpoint every-n functions (default: 1)
        flatten (bool): flatten nn.Sequential of nn.Sequentials
        skip_last (bool): skip checkpointing the last function in the sequence if True
        preserve_rng_state (bool, optional, default=True):  Omit stashing and restoring
            the RNG state during each checkpoint.

    Returns:
        Output of running :attr:`functions` sequentially on :attr:`*inputs`

    Example:
        >>> model = nn.Sequential(...)
        >>> input_var = checkpoint_seq(model, input_var, every=2)
    """

    def run_function(start, end, functions):
        def forward(_x):
            for j in range(start, end + 1):
                _x = functions[j](_x)
            return _x

        return forward

    if isinstance(functions, torch.nn.Sequential):
        functions = functions.children()
    if flatten:
        functions = chain.from_iterable(functions)
    if not isinstance(functions, (tuple, list)):
        functions = tuple(functions)

    num_checkpointed = len(functions)
    if skip_last:
        num_checkpointed -= 1
    end = -1
    for start in range(0, num_checkpointed, every):
        end = min(start + every - 1, num_checkpointed - 1)
        x = checkpoint(
            run_function(start, end, functions),
            x,
            preserve_rng_state=preserve_rng_state,
        )
    if skip_last:
        return run_function(end + 1, len(functions) - 1, functions)(x)
    return x



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import numpy as np


class DomainNorm2(nn.Module):
    def __init__(self, channel, l2=True):
        super(DomainNorm, self).__init__()
        self.normalize = nn.InstanceNorm2d(num_features=channel, affine=False)
        self.l2 = l2
        self.weight = nn.Parameter(torch.ones(1, channel, 1, 1))
        self.bias = nn.Parameter(torch.zeros(1, channel, 1, 1))
        self.weight.requires_grad = True
        self.bias.requires_grad = True

    def forward(self, x):
        x = self.normalize(x)
        if self.l2:
            x = F.normalize(x, p=2, dim=1)
        return x * self.weight + self.bias


class DomainNorm(nn.Module):
    def __init__(self, channel, l2=True):
        super(DomainNorm, self).__init__()
        self.normalize = nn.InstanceNorm2d(num_features=channel, affine=True)
        self.l2 = l2

    def forward(self, x):
        if self.l2:
            x = F.normalize(x, p=2, dim=1)
        x = self.normalize(x)
        return x


class BasicConv(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        deconv=False,
        is_3d=False,
        bn=True,
        l2=True,
        relu=True,
        **kwargs,
    ):
        super(BasicConv, self).__init__()
        #        print(in_channels, out_channels, deconv, is_3d, bn, relu, kwargs)
        self.relu = relu
        self.use_bn = bn
        self.l2 = l2
        if is_3d:
            if deconv:
                self.conv = nn.ConvTranspose3d(
                    in_channels, out_channels, bias=False, **kwargs
                )
            else:
                self.conv = nn.Conv3d(in_channels, out_channels, bias=False, **kwargs)
            self.bn = nn.BatchNorm3d(out_channels)
        else:
            if deconv:
                self.conv = nn.ConvTranspose2d(
                    in_channels, out_channels, bias=False, **kwargs
                )
            else:
                self.conv = nn.Conv2d(in_channels, out_channels, bias=False, **kwargs)
            self.bn = DomainNorm(channel=out_channels, l2=self.l2)

    #            self.bn = nn.InstanceNorm2d(out_channels)

    def forward(self, x):
        x = self.conv(x)
        if self.use_bn:
            x = self.bn(x)
        if self.relu:
            x = F.relu(x, inplace=True)
        return x


class Conv2x(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        deconv=False,
        is_3d=False,
        concat=True,
        bn=True,
        relu=True,
        kernel=None,
    ):
        super(Conv2x, self).__init__()
        self.concat = concat
        if kernel is not None:
            self.kernel = kernel
        # elif deconv and is_3d:
        #    kernel = (4, 4, 4)
        elif deconv:
            kernel = 4
        else:
            kernel = 3
        self.conv1 = BasicConv(
            in_channels,
            out_channels,
            deconv,
            is_3d,
            bn=True,
            relu=True,
            kernel_size=kernel,
            stride=2,
            padding=1,
        )

        if self.concat:
            self.conv2 = BasicConv(
                out_channels * 2,
                out_channels,
                False,
                is_3d,
                bn,
                relu,
                kernel_size=3,
                stride=1,
                padding=1,
            )
        else:
            self.conv2 = BasicConv(
                out_channels,
                out_channels,
                False,
                is_3d,
                bn,
                relu,
                kernel_size=3,
                stride=1,
                padding=1,
            )

    def forward(self, x, rem):
        x = self.conv1(x)
        # print(x.shape, rem.shape)
        assert x.size() == rem.size(), [x.size(), rem.size()]
        if self.concat:
            x = torch.cat((x, rem), 1)
        else:
            x = x + rem
        x = self.conv2(x)
        return x


class SGABlock(nn.Module):
    def __init__(self, channels=32, refine=False):
        super(SGABlock, self).__init__()
        self.refine = refine
        if self.refine:
            self.bn_relu = nn.Sequential(
                nn.BatchNorm3d(channels), nn.ReLU(inplace=True)
            )
            self.conv_refine = BasicConv(
                channels, channels, is_3d=True, kernel_size=3, padding=1, relu=False
            )
        #            self.conv_refine1 = BasicConv(8, 8, is_3d=True, kernel_size=1, padding=1)
        else:
            self.bn = nn.BatchNorm3d(channels)

        try:
            from .libs.GANet.modules.GANet import SGA
        except ImportError:
            raise ImportError(
                "ERROR: Could not import libs.GANet required for SeparableFlow."
                + " Go to ptlflow/models/separableflow/ and then install GANet by running: bash compile.sh"
            )
        self.SGA = SGA()
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x, g):
        rem = x
        k1, k2, k3, k4 = torch.split(g, (5, 5, 5, 5), 1)
        k1 = F.normalize(k1, p=1, dim=1)
        k2 = F.normalize(k2, p=1, dim=1)
        k3 = F.normalize(k3, p=1, dim=1)
        k4 = F.normalize(k4, p=1, dim=1)
        x = self.SGA(x, k1, k2, k3, k4)
        if self.refine:
            x = self.bn_relu(x)
            x = self.conv_refine(x)
        else:
            x = self.bn(x)
        assert x.size() == rem.size()
        x += rem
        return self.relu(x)


def freeze_layers(model):
    for child in model.children():
        for param in child.parameters():
            param.requires_grad = False
    for param in model.parameters():
        param.requires_grad = False


class ShiftRegression(nn.Module):
    def __init__(self, max_shift=192):
        super(ShiftRegression, self).__init__()
        self.max_shift = max_shift

    #        self.disp = Variable(torch.Tensor(np.reshape(np.array(range(self.maxdisp)),[1,self.maxdisp,1,1])).cuda(), requires_grad=False)

    def forward(self, x, max_shift=None):
        if max_shift is not None:
            self.max_shift = max_shift
        assert x.is_contiguous() == True
        with torch.cuda.device_of(x):
            shift = Variable(
                torch.Tensor(
                    np.reshape(
                        np.array(range(-self.max_shift, self.max_shift + 1)),
                        [1, self.max_shift * 2 + 1, 1, 1],
                    )
                ).cuda(),
                requires_grad=False,
            )
            shift = shift.repeat(x.size()[0], 1, x.size()[2], x.size()[3])
            out = torch.sum(x * shift, dim=1, keepdim=True)
        return out


class ShiftEstimate(nn.Module):
    def __init__(self, max_shift=192, InChannel=24):
        super(ShiftEstimate, self).__init__()
        self.max_shift = int(max_shift / 2)
        self.softmax = nn.Softmax(dim=1)
        self.regression = ShiftRegression(max_shift=self.max_shift + 1)
        self.conv3d_2d = nn.Conv3d(
            InChannel, 1, (3, 3, 3), (1, 1, 1), (1, 1, 1), bias=True
        )
        # self.upsample_cost = FilterUpsample()

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/3, W/3, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 3, 3, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(3 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 1, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 1, 3 * H, 3 * W)

    def forward(self, x):
        # N, _,  _, H, W = g.size()
        # assert (x.size(3)==H and x.size(4)==W)
        # x = self.upsample_cost(x, g)
        # print(x.size(), g.size())
        x = F.interpolate(
            self.conv3d_2d(x),
            [self.max_shift * 2 + 1, x.size()[3] * 4, x.size()[4] * 4],
            mode="trilinear",
            align_corners=True,
        )
        x = torch.squeeze(x, 1)
        x = self.softmax(x)
        x = self.regression(x)
        x = F.interpolate(
            x, [x.size()[2] * 2, x.size()[3] * 2], mode="bilinear", align_corners=True
        )
        return x * 2


class ShiftEstimate2(nn.Module):
    def __init__(self, max_shift=100, InChannel=24):
        super(ShiftEstimate2, self).__init__()
        self.max_shift = int(max_shift // 4)
        self.softmax = nn.Softmax(dim=1)
        self.regression = ShiftRegression()
        self.conv3d_2d = nn.Conv3d(
            InChannel, 1, kernel_size=3, stride=1, padding=1, bias=True
        )
        # self.upsample_cost = FilterUpsample()

    def forward(self, x, max_shift=None):
        if max_shift is not None:
            assert (max_shift // 8 * 2 + 1) == x.shape[2], [
                x.shape,
                max_shift,
                max_shift // 8 * 2 + 1,
            ]
            # assert(x.size() == rem.size()),[x.size(), rem.size()]
            self.max_shift = max_shift // 4
        x = F.interpolate(
            self.conv3d_2d(x),
            [self.max_shift * 2 + 1, x.size()[3] * 2, x.size()[4] * 2],
            mode="trilinear",
            align_corners=True,
        )
        #        x = self.conv3d_2d(x)
        x = torch.squeeze(x, 1)

        x = self.softmax(x)
        x = self.regression(x, self.max_shift)
        x = F.interpolate(
            x, [x.size()[2] * 4, x.size()[3] * 4], mode="bilinear", align_corners=True
        )

        return x * 4


class CostAggregation(nn.Module):
    def __init__(self, max_shift=400, in_channel=8):
        super(CostAggregation, self).__init__()
        self.max_shift = max_shift
        self.in_channel = in_channel  # t(self.max_shift / 6) * 2 + 1
        self.inner_channel = 8
        self.conv0 = BasicConv(
            self.in_channel,
            self.inner_channel,
            is_3d=True,
            kernel_size=3,
            padding=1,
            relu=True,
        )

        self.conv1a = BasicConv(
            self.inner_channel,
            self.inner_channel * 2,
            is_3d=True,
            kernel_size=3,
            stride=2,
            padding=1,
        )
        self.conv2a = BasicConv(
            self.inner_channel * 2,
            self.inner_channel * 4,
            is_3d=True,
            kernel_size=3,
            stride=2,
            padding=1,
        )
        self.conv3a = BasicConv(
            self.inner_channel * 4,
            self.inner_channel * 6,
            is_3d=True,
            kernel_size=3,
            stride=2,
            padding=1,
        )

        self.deconv1a = Conv2x(
            self.inner_channel * 2,
            self.inner_channel,
            deconv=True,
            is_3d=True,
            relu=True,
        )
        self.deconv2a = Conv2x(
            self.inner_channel * 4, self.inner_channel * 2, deconv=True, is_3d=True
        )
        self.deconv3a = Conv2x(
            self.inner_channel * 6, self.inner_channel * 4, deconv=True, is_3d=True
        )

        self.conv1b = BasicConv(
            self.inner_channel,
            self.inner_channel * 2,
            is_3d=True,
            kernel_size=3,
            stride=2,
            padding=1,
        )
        self.conv2b = BasicConv(
            self.inner_channel * 2,
            self.inner_channel * 4,
            is_3d=True,
            kernel_size=3,
            stride=2,
            padding=1,
        )
        self.conv3b = BasicConv(
            self.inner_channel * 4,
            self.inner_channel * 6,
            is_3d=True,
            kernel_size=3,
            stride=2,
            padding=1,
        )

        self.deconv1b = Conv2x(
            self.inner_channel * 2,
            self.inner_channel,
            deconv=True,
            is_3d=True,
            relu=True,
            kernel=(3, 4, 4),
        )
        self.deconv2b = Conv2x(
            self.inner_channel * 4,
            self.inner_channel * 2,
            deconv=True,
            is_3d=True,
            kernel=(3, 4, 4),
        )
        self.deconv3b = Conv2x(
            self.inner_channel * 6,
            self.inner_channel * 4,
            deconv=True,
            is_3d=True,
            kernel=(3, 4, 4),
        )
        self.shift0 = ShiftEstimate2(self.max_shift, self.inner_channel)
        self.shift1 = ShiftEstimate2(self.max_shift, self.inner_channel)
        self.shift2 = ShiftEstimate2(self.max_shift, self.inner_channel)
        self.sga1 = SGABlock(channels=self.inner_channel, refine=True)
        self.sga2 = SGABlock(channels=self.inner_channel, refine=True)
        self.sga3 = SGABlock(channels=self.inner_channel, refine=True)
        self.sga11 = SGABlock(channels=self.inner_channel * 2, refine=True)
        self.sga12 = SGABlock(channels=self.inner_channel * 2, refine=True)
        self.corr_output = BasicConv(
            self.inner_channel, 1, is_3d=True, kernel_size=3, padding=1, relu=False
        )
        self.corr2cost = Corr2Cost()

    def forward(self, x, g, max_shift=400, is_ux=True):
        x = self.conv0(x)
        x = self.sga1(x, g["sg1"])
        rem0 = x

        if self.training:
            cost = self.corr2cost(x, max_shift // 8, is_ux)
            shift0 = self.shift0(cost, max_shift)

        x = self.conv1a(x)
        x = self.sga11(x, g["sg11"])
        rem1 = x
        x = self.conv2a(x)
        rem2 = x
        x = self.conv3a(x)
        rem3 = x

        x = self.deconv3a(x, rem2)
        rem2 = x
        x = self.deconv2a(x, rem1)
        x = self.sga12(x, g["sg12"])
        rem1 = x
        x = self.deconv1a(x, rem0)
        x = self.sga2(x, g["sg2"])
        rem0 = x
        cost = self.corr2cost(x, max_shift // 8, is_ux)
        if self.training:
            shift1 = self.shift1(cost, max_shift)
        corr = self.corr_output(x)
        rem0 = cost
        x = self.conv1b(cost)
        rem1 = x
        x = self.conv2b(x)
        rem2 = x
        x = self.conv3b(x)
        x = self.deconv3b(x, rem2)
        x = self.deconv2b(x, rem1)
        x = self.deconv1b(x, rem0)
        x = self.sga3(x, g["sg3"])
        shift2 = self.shift2(x, max_shift)
        if self.training:
            return shift0, shift1, shift2, corr
        else:
            return shift2, corr


class Corr2Cost(nn.Module):
    def __init__(self):
        super(Corr2Cost, self).__init__()

    def coords_grid(self, batch, ht, wd, device):
        coords = torch.meshgrid(
            torch.arange(ht, device=device),
            torch.arange(wd, device=device),
            indexing="ij",
        )
        coords = torch.stack(coords[::-1], dim=0).float()
        return coords[None].repeat(batch, 1, 1, 1)

    def bilinear_sampler(self, img, coords, mode="bilinear", mask=False):
        """Wrapper for grid_sample, uses pixel coordinates"""
        H, W = img.shape[-2:]
        xgrid, ygrid = coords.split([1, 1], dim=-1)
        xgrid = 2 * xgrid / (W - 1) - 1
        assert torch.unique(ygrid).numel() == 1 and H == 1  # This is a stereo problem

        grid = torch.cat([xgrid, ygrid], dim=-1)
        img = F.grid_sample(img, grid, align_corners=True)

        if mask:
            mask = (xgrid > -1) & (ygrid > -1) & (xgrid < 1) & (ygrid < 1)
            return img, mask.float()

        return img

    def forward(self, corr, maxdisp=50, is_ux=True):
        batch, dim, d, h, w = corr.shape
        corr = corr.permute(0, 3, 4, 1, 2).reshape(batch * h * w, dim, 1, d)
        with torch.no_grad():
            coords = self.coords_grid(batch, h, w, corr.device)
            if is_ux:
                coords = coords[:, :1, :, :]
            else:
                coords = coords[:, 1:, :, :]
            dx = torch.linspace(-maxdisp, maxdisp, maxdisp * 2 + 1)
            dx = dx.view(1, 1, 2 * maxdisp + 1, 1).to(corr.device)
            x0 = dx + coords.reshape(batch * h * w, 1, 1, 1)
            y0 = torch.zeros_like(x0)
            # if is_ux:
            coords_lvl = torch.cat([x0, y0], dim=-1)
        # else:
        #     coords_lvl = torch.cat([y0, x0], dim=-1)
        corr = self.bilinear_sampler(corr, coords_lvl)
        # print(corr.shape)
        corr = corr.view(batch, h, w, dim, maxdisp * 2 + 1)
        corr = corr.permute(0, 3, 4, 1, 2).contiguous().float()
        return corr



import torch


class ConvBlock(torch.nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, stride, padding):
        super(ConvBlock, self).__init__()

        self.conv = torch.nn.Conv2d(
            in_planes,
            out_planes,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            padding_mode="zeros",
            bias=False,
        )
        self.relu = torch.nn.LeakyReLU(negative_slope=0.1, inplace=False)

    def forward(self, x):
        return self.relu(self.conv(x))


class Refine(torch.nn.Module):
    def __init__(
        self, context_dim, iter_context_dim, num_layers, levels, radius, inter_dim
    ):
        super(Refine, self).__init__()

        self.radius = radius

        self.conv1 = ConvBlock(
            (radius * 2 + 1) ** 2 * levels + context_dim + iter_context_dim + 2 + 1,
            context_dim + iter_context_dim,
            kernel_size=3,
            stride=1,
            padding=1,
        )

        self.conv2 = ConvBlock(
            context_dim + iter_context_dim,
            inter_dim,
            kernel_size=3,
            stride=1,
            padding=1,
        )

        self.conv_layers = torch.nn.ModuleList(
            [
                ConvBlock(inter_dim, inter_dim, kernel_size=3, stride=1, padding=1)
                for i in range(num_layers)
            ]
        )

        self.conv3 = torch.nn.Conv2d(
            inter_dim,
            iter_context_dim + 2,
            kernel_size=3,
            stride=1,
            padding=1,
            padding_mode="zeros",
            bias=True,
        )

        # self.hidden_act = torch.nn.Tanh()
        self.hidden_act = torch.nn.Hardtanh(min_val=-4.0, max_val=4.0)
        # self.hidden_norm = torch.nn.BatchNorm2d(feature_dim)

    def init_bhwd(self, batch_size, height, width, device, amp):
        self.radius_emb = (
            torch.tensor(
                self.radius, dtype=torch.half if amp else torch.float, device=device
            )
            .view(1, -1, 1, 1)
            .expand([batch_size, 1, height, width])
        )

    def forward(self, corrs, context, iter_context, flow0):
        x = torch.cat([corrs, context, iter_context, flow0, self.radius_emb], dim=1)

        x = self.conv1(x)

        x = self.conv2(x)

        for layer in self.conv_layers:
            x = layer(x)

        x = self.conv3(x)

        return self.hidden_act(x[:, 2:]), x[:, :2]



""" Normalization layers and wrappers

Norm layer definitions that support fast norm and consistent channel arg order (always first arg).

Hacked together by / Copyright 2022 Ross Wightman
"""

import torch
import torch.nn as nn
import torch.nn.functional as F


class LayerNorm(nn.LayerNorm):
    """LayerNorm w/ fast norm option"""

    def __init__(self, num_channels, eps=1e-6, affine=True):
        super().__init__(num_channels, eps=eps, elementwise_affine=affine)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        return x


class LayerNorm2d(nn.LayerNorm):
    """LayerNorm for channels of '2D' spatial NCHW tensors"""

    def __init__(self, num_channels, eps=1e-6, affine=True):
        super().__init__(num_channels, eps=eps, elementwise_affine=affine)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.permute(0, 2, 3, 1)
        x = F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        x = x.permute(0, 3, 1, 2)
        return x



import torch
import torch.nn as nn

from ..functions.quadtree_attention import score_computation_op, value_aggregation_op
from einops.einops import rearrange


class QTAttA(nn.Module):
    def __init__(
        self,
        nhead,
        dim,
        topks=[32, 32, 32, 32],
        scale=None,
        use_dropout=False,
        attention_dropout=0.1,
    ):
        super().__init__()
        self.use_dropout = use_dropout
        self.topks = topks
        self.nhead = nhead
        self.dim = dim

    def process_coarse_level(self, query, key, value, topk):
        bs, c, h, w = key.shape
        cur_dim = key.shape[1] // self.nhead

        key = rearrange(key, "b c h w -> b (h w) c").view(
            bs, -1, self.nhead, cur_dim
        )  # [N, S, H, D]
        value = rearrange(value, "b c h w -> b (h w) c").view(
            bs, -1, self.nhead, cur_dim
        )  # [N, S, H, D]
        query = rearrange(query, "b c h w -> b (h w) c").view(
            bs, -1, self.nhead, cur_dim
        )

        QK = torch.einsum("nlhd,nshd->nlsh", query, key)
        softmax_temp = 1.0 / cur_dim**0.5  # sqrt(D)
        A = torch.softmax(softmax_temp * QK, dim=-2)

        # mask out top K tokens
        topk_score, topk_idx = torch.topk(A, dim=-2, k=topk, largest=True)
        mask = torch.ones_like(A)
        mask = mask.scatter(
            dim=-2, index=topk_idx, src=torch.zeros_like(topk_idx).float()
        )

        # message is only computed within the unmasked
        message = torch.einsum(
            "nlsh,nshd->nlhd", A * mask, value
        )  # .reshape(bs, h, w, self.nhead, cur_dim)

        return A, message, topk_score, topk_idx

    def process_fine_level(
        self, query, key, value, topk_score, topk_pos, topk_prev, topk, final=False
    ):
        bs, c, h, w = key.shape

        cur_dim = key.shape[1] // self.nhead
        key = rearrange(key, "b c h w -> b (h w) c").view(
            bs, -1, self.nhead, cur_dim
        )  # [N, S, H, D]
        value = rearrange(value, "b c h w -> b (h w) c").view(
            bs, -1, self.nhead, cur_dim
        )  # [N, S, H, D]

        query = query.view(bs, c, h // 2, 2, w // 2, 2)
        query = rearrange(query, "b c h t1 w t2-> b (h w) (t1 t2) c ").view(
            bs, -1, 4, self.nhead, cur_dim
        )

        # convert 2d coordinates to 1d index
        idx_gather = []
        topk_pos = topk_pos * 2
        for x in [0, 1]:
            for y in [0, 1]:
                idx = (topk_pos[0] + x) * w + topk_pos[1] + y  # convert to index
                idx_gather.append(idx)

        idx = torch.stack(idx_gather, dim=3)  # [N, L, K, 4, H, D]

        # Compute score
        # query: [b, N, 4, H, D]
        # key: [b, 4N, H, D]
        # idx: [b, N, K, 4, H]
        # QK: [b, N, 4, 4K, H]
        QK = score_computation_op(
            query, key.contiguous(), idx.view(bs, -1, topk_prev * 4, self.nhead)
        )
        QK = rearrange(QK, "n l w (k f) h -> n l w k f h", k=topk_prev, f=4)
        softmax_temp = 1.0 / cur_dim**0.5  # sqrt(D)
        A = torch.softmax(softmax_temp * QK, dim=-2)  # [N, L//scale**i, K, 4, H]
        # Score redistribution
        topk_score = topk_score.unsqueeze(-2).unsqueeze(2)
        A = (A * topk_score).reshape(bs, -1, 4, topk_prev * 4, self.nhead)
        idx = idx.view(bs, -1, 1, topk_prev * 4, self.nhead).repeat(
            1, 1, 4, 1, 1
        )  # [N, L,4, K*4, H]
        topk_score, topk_idx = torch.topk(A, dim=-2, k=topk, largest=True)

        if not final:
            mask = torch.ones_like(A)
            mask = mask.scatter(
                dim=-2, index=topk_idx, src=torch.zeros_like(topk_idx).float()
            )
            message = value_aggregation_op(A * mask, value.contiguous(), idx)
        else:
            message = value_aggregation_op(A, value.contiguous(), idx)

        if not final:
            topk_idx = torch.gather(idx, index=topk_idx, dim=-2)
            topk_idx = rearrange(
                topk_idx, "b (h w) (t1 t2) k nh -> b (h t1 w t2) k nh", h=h // 2, t1=2
            )  # reshape back
            topk_score = rearrange(
                topk_score, "b (h w) (t1 t2) k nh -> b (h t1 w t2) k nh", h=h // 2, t1=2
            )  # reshape back

        return A, message, topk_score, topk_idx

    def forward(self, queries, keys, values, q_mask=None, kv_mask=None):
        """Multi-head quadtree attention
        Args:
            queries: Query pyramid [N, C, H, W]
            keys: Key pyramid [N, C, H, W]
            values: Value pyramid [N, C, H, W]
        Returns:
            message: (N, C, H, W)
        """

        bs = queries[0].shape[0]
        messages = []
        topk = self.topks[0]

        for i, (query, key, value) in enumerate(
            zip(reversed(queries), reversed(keys), reversed(values))
        ):
            bs, c, h, w = key.shape
            if i == 0:
                A, message, topk_score, topk_idx = self.process_coarse_level(
                    query, key, value, topk
                )  # Full attention for coarest level
            else:
                topk_prev = topk
                topk = self.topks[i]
                final = True if i == len(queries) - 1 else False
                A, message, topk_score, topk_idx = self.process_fine_level(
                    query, key, value, topk_score, topk_pos, topk_prev, topk, final
                )  # Quadtree attention

            messages.append(message)
            if topk_idx is not None:
                topk_pos = torch.stack(
                    [topk_idx // w, topk_idx % w]
                )  # convert to coordinate

        final_message = 0
        for i, m in enumerate(messages):
            if i == 0:
                final_message = m
            else:
                final_message = final_message.unsqueeze(2) + m
                final_message = rearrange(
                    final_message,
                    "b (H W) (t1 t2) h d -> b (H t1 W t2) h d",
                    t1=2,
                    t2=2,
                    H=queries[-i].shape[2],
                )

        return final_message


class QTAttB(nn.Module):
    def __init__(
        self,
        nhead,
        dim,
        scale,
        topks=[32, 32, 32, 32],
        use_dropout=False,
        attention_dropout=0.1,
        lepe=False,
    ):
        super().__init__()
        self.use_dropout = use_dropout
        self.topks = topks
        self.nhead = nhead
        self.dim = dim
        self.lepe = lepe
        if lepe:  # locally enhanced position encoding
            self.get_vs = nn.ModuleList(
                [
                    nn.Conv2d(
                        dim * nhead,
                        dim * nhead,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                        groups=dim * nhead,
                    )
                    for _ in range(scale)
                ]
            )
        self.register_parameter("weight", nn.Parameter(torch.randn(scale)))

    def process_coarse_level(self, query, key, value, topk):
        bs, c, h, w = key.shape

        cur_dim = key.shape[1] // self.nhead
        key = rearrange(key, "b c h w -> b (h w) c").view(
            bs, -1, self.nhead, cur_dim
        )  # [N, S, H, D]
        value = rearrange(value, "b c h w -> b (h w) c").view(
            bs, -1, self.nhead, cur_dim
        )  # [N, S, H, D]
        query = rearrange(query, "b c h w -> b (h w) c").view(
            bs, -1, self.nhead, cur_dim
        )
        QK = torch.einsum("nlhd,nshd->nlsh", query, key)
        softmax_temp = 1.0 / cur_dim**0.5  # sqrt(D)

        A = torch.softmax(softmax_temp * QK, dim=-2)
        topk_score, topk_idx = torch.topk(A, dim=-2, k=topk, largest=True)

        message = torch.einsum(
            "nlsh,nshd->nlhd", A, value
        )  # .reshape(bs, h, w, self.nhead, cur_dim)

        return A, message, topk_score, topk_idx

    def process_fine_level(
        self, query, key, value, topk_score, topk_pos, topk_prev, topk, final=False
    ):
        bs, c, h, w = key.shape

        cur_dim = key.shape[1] // self.nhead
        key = rearrange(key, "b c h w -> b (h w) c").view(
            bs, -1, self.nhead, cur_dim
        )  # [N, S, H, D]
        value = rearrange(value, "b c h w -> b (h w) c").view(
            bs, -1, self.nhead, cur_dim
        )  # [N, S, H, D]

        query = query.view(bs, c, h // 2, 2, w // 2, 2)
        query = rearrange(query, "b c h t1 w t2-> b (h w) (t1 t2) c ").view(
            bs, -1, 4, self.nhead, cur_dim
        )

        # convert 2D coordiantes to 1D index
        topk_pos = topk_pos * 2
        idx_gather = []
        for x in [0, 1]:
            for y in [0, 1]:
                idx = (topk_pos[0] + x) * w + topk_pos[1] + y  # convert to index
                idx_gather.append(idx)
        idx = torch.stack(idx_gather, dim=3)  # [N, L, K, 4, H, D]

        # score computation
        # query: [b, N, 4, H, D]
        # key: [b, 4N, H, D]
        # idx: [b, N, K, 4, H]
        # QK: [b, N, 4, 4K, H]
        QK = score_computation_op(
            query, key.contiguous(), idx.view(bs, -1, topk_prev * 4, self.nhead)
        )
        softmax_temp = 1.0 / cur_dim**0.5  # sqrt(D)
        A = torch.softmax(softmax_temp * QK, dim=-2)  # [N, L//scale**i, K, 4, H]
        A = A.reshape(bs, -1, 4, topk_prev * 4, self.nhead)
        idx = idx.view(bs, -1, 1, topk_prev * 4, self.nhead).repeat(
            1, 1, 4, 1, 1
        )  # [N, L,4, K*4, H]

        topk_score, topk_idx = torch.topk(A, dim=-2, k=topk, largest=True)
        message = value_aggregation_op(A, value.contiguous(), idx)
        topk_idx = torch.gather(idx, index=topk_idx, dim=-2)
        topk_idx = rearrange(
            topk_idx, "b (h w) (t1 t2) k nh -> b (h t1 w t2) k nh", h=h // 2, t1=2
        )  # reshape back
        topk_score = rearrange(
            topk_score, "b (h w) (t1 t2) k nh -> b (h t1 w t2) k nh", h=h // 2, t1=2
        )  # reshape back

        return A, message, topk_score, topk_idx

    def forward(self, queries, keys, values, q_mask=None, kv_mask=None):
        """Multi-head quadtree attention
        Args:
            queries: Query pyramid [N, C, H, W]
            keys: Key pyramid [N, C, H, W]
            values: Value pyramid [N, C, H, W]
        Returns:
            message: (N, C, H, W)
        """

        bs = queries[0].shape[0]

        messages = []
        topk = self.topks[0]
        for i, (query, key, value) in enumerate(
            zip(reversed(queries), reversed(keys), reversed(values))
        ):
            bs, c, h, w = key.shape
            if i == 0:  # Full attention for the coarest level
                A, message, topk_score, topk_idx = self.process_coarse_level(
                    query, key, value, topk
                )
            else:
                topk_prev = topk
                topk = self.topks[i]
                final = True if i == len(queries) - 1 else False
                A, message, topk_score, topk_idx = self.process_fine_level(
                    query, key, value, topk_score, topk_pos, topk_prev, topk, final
                )

            messages.append(message)
            topk_pos = torch.stack(
                [topk_idx // w, topk_idx % w]
            )  # convert to coordinate

        # Merge messages of different layers
        final_message = 0

        weight = torch.softmax(self.weight, dim=0)
        for i, m in enumerate(messages):
            if self.lepe:
                H, W = values[-(i + 1)].shape[-2:]
                lepe = self.get_vs[i](values[-(i + 1)])

            if i == 0:
                if self.lepe:
                    lepe = rearrange(
                        lepe, "b (hd d) H W -> b (H W) hd d", hd=self.nhead
                    )
                    final_message = (m + lepe) * weight[i]
                else:
                    final_message = m * weight[i]
            else:
                if self.lepe:
                    lepe = rearrange(
                        lepe,
                        "b (hd d) (H t1) (W t2) -> b (H W) (t1 t2) hd d",
                        hd=self.nhead,
                        t1=2,
                        t2=2,
                    )
                    final_message = final_message.unsqueeze(2) + (m + lepe) * weight[i]
                else:
                    final_message = final_message.unsqueeze(2) + m * weight[i]

                final_message = rearrange(
                    final_message,
                    "b (H W) (t1 t2) h d -> b (H t1 W t2) h d",
                    t1=2,
                    t2=2,
                    H=queries[-i].shape[2],
                )
        return final_message



import torch
import torch.nn.functional as F


class Correlation1D:
    def __init__(
        self,
        feature1,
        feature2,
        radius=32,
        x_correlation=False,
    ):
        self.radius = radius
        self.x_correlation = x_correlation

        if self.x_correlation:
            self.corr = self.corr_x(feature1, feature2)  # [B*H*W, 1, 1, W]
        else:
            self.corr = self.corr_y(feature1, feature2)  # [B*H*W, 1, H, 1]

    def __call__(self, coords):
        r = self.radius
        coords = coords.permute(0, 2, 3, 1)  # [B, H, W, 2]
        b, h, w = coords.shape[:3]

        if self.x_correlation:
            dx = torch.linspace(-r, r, 2 * r + 1)
            dy = torch.zeros_like(dx)
            delta_x = torch.stack((dx, dy), dim=-1).to(
                device=coords.device, dtype=coords.dtype
            )  # [2r+1, 2]

            coords_x = coords[:, :, :, 0]  # [B, H, W]
            coords_x = torch.stack(
                (coords_x, torch.zeros_like(coords_x)), dim=-1
            )  # [B, H, W, 2]

            centroid_x = coords_x.view(b * h * w, 1, 1, 2)  # [B*H*W, 1, 1, 2]
            coords_x = centroid_x + delta_x  # [B*H*W, 1, 2r+1, 2]

            coords_x = 2 * coords_x / (w - 1) - 1  # [-1, 1], y is always 0

            corr_x = F.grid_sample(
                self.corr, coords_x, mode="bilinear", align_corners=True
            )  # [B*H*W, G, 1, 2r+1]

            corr_x = corr_x.view(b, h, w, -1)  # [B, H, W, (2r+1)*G]
            corr_x = corr_x.permute(0, 3, 1, 2).contiguous()  # [B, (2r+1)*G, H, W]
            return corr_x
        else:  # y correlation
            dy = torch.linspace(-r, r, 2 * r + 1).to(
                device=coords.device, dtype=coords.dtype
            )
            dx = torch.zeros_like(dy)
            delta_y = torch.stack((dx, dy), dim=-1).to(coords.device)  # [2r+1, 2]
            delta_y = delta_y.unsqueeze(1).unsqueeze(0)  # [1, 2r+1, 1, 2]

            coords_y = coords[:, :, :, 1]  # [B, H, W]
            coords_y = torch.stack(
                (torch.zeros_like(coords_y), coords_y), dim=-1
            )  # [B, H, W, 2]

            centroid_y = coords_y.view(b * h * w, 1, 1, 2)  # [B*H*W, 1, 1, 2]
            coords_y = centroid_y + delta_y  # [B*H*W, 2r+1, 1, 2]

            coords_y = 2 * coords_y / (h - 1) - 1  # [-1, 1], x is always 0

            corr_y = F.grid_sample(
                self.corr, coords_y, mode="bilinear", align_corners=True
            )  # [B*H*W, G, 2r+1, 1]

            corr_y = corr_y.view(b, h, w, -1)  # [B, H, W, (2r+1)*G]
            corr_y = corr_y.permute(0, 3, 1, 2).contiguous()  # [B, (2r+1)*G, H, W]

            return corr_y

    def corr_x(self, feature1, feature2):
        b, c, h, w = feature1.shape  # [B, C, H, W]
        scale_factor = c**0.5

        # x direction
        feature1 = feature1.permute(0, 2, 3, 1)  # [B, H, W, C]
        feature2 = feature2.permute(0, 2, 1, 3)  # [B, H, C, W]
        corr = torch.matmul(feature1, feature2)  # [B, H, W, W]

        corr = corr.unsqueeze(3).unsqueeze(3)  # [B, H, W, 1, 1, W]
        corr = corr / scale_factor
        corr = corr.flatten(0, 2)  # [B*H*W, 1, 1, W]

        return corr

    def corr_y(self, feature1, feature2):
        b, c, h, w = feature1.shape  # [B, C, H, W]
        scale_factor = c**0.5

        # y direction
        feature1 = feature1.permute(0, 3, 2, 1)  # [B, W, H, C]
        feature2 = feature2.permute(0, 3, 1, 2)  # [B, W, C, H]
        corr = torch.matmul(feature1, feature2)  # [B, W, H, H]

        corr = (
            corr.permute(0, 2, 1, 3).contiguous().view(b, h, w, 1, h, 1)
        )  # [B, H, W, 1, H, 1]
        corr = corr / scale_factor
        corr = corr.flatten(0, 2)  # [B*H*W, 1, H, 1]

        return corr
